/***************************************************************************
    rs4decoderas.c  -  Reed-Solomon & Burst Error Protection
    extended   with use of erasure positions by PA0MBO
                             -------------------
    begin                : Mon Jul  30 20:27:09 MEST 2001
    copyright            : (C) 2000 by Guido Fiala
    email                : gfiala@s.netic.de
	ties xxxxx
 ***************************************************************************/

/***************************************************************************
    Slightly modified by ttsiodras at removethis gmail dot com

    It doesn't write the 3 parameters of rsbep as a single line before the 
    data, as this makes the stream fragile (if this information is lost, 
    decoding fails)

    It uses a default value of 16*255=4080 for parameter D, and it can thus 
    tolerate 4080*16=65280 consecutive bytes to be lost anywhere in the 
    stream, and still recover...
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***********************************************************************
 *
 * Changed for decode only and much simpler command lines for use
 * in rxamadrm by Ties Bos - PA0MBO
 * Date Dec 20th 2011
 * and use of erasure positions   on May 15th 2012
 * **********************************************************************/

 
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <assert.h>

  /* plain & slow C version (can also be used for anything different from RS(255,223)) */
#include "rs4.h"       
#define rse32 encode_rs       
#define rsd32 eras_dec_rs

#define ERR_BURST_LEN (255*16)   
#define RS_BSIZE 255            
#define RS_DSIZE 128            
#define RSBEP_VERSION "0.0.3"
#define RSBEP_HDR "rsbep"
#define MAX_EXPECTED_CHUNKS 1000  /* reeds about 300k sized files */

/* just a "large" random string because it's so unlikely to be in the data by chance - better idea? */
#define RSBEP_MAGIC_NUM "Inʘ8R7S"

#define PHDR_LEN  54    /* (40+strlen(RSBEP_MAGIC_NUM)) //maximum expected len of header line... */


union long_byte_union
{
        long long i;
        char b[8];
};

/* this function is used to copy src to dst while separating consecutive bytes
equally distributed over the entire array, it tries to utilize several
cpu-pipes and as high as possible memory efficiency without using asm.
(but not several CPU's...anyway -it's only 15% of CPU-time, most is RS) */

void distribute(dtype *src, dtype *dst, int rows, int cols, int reverse)
{
        register long i,j,rc,ri,rl;
        union long_byte_union lb[8];
        rc=rows*cols;
        ri=0;
        rl = reverse ? cols : rows;
        for(i=0;i<rc;i+=64) /* we read 64 byte at once  */
        {
                lb[0].i=*((long long*)(src+i));    
                lb[1].i=*((long long*)(src+i+8));  
                lb[2].i=*((long long*)(src+i+16)); 
                lb[3].i=*((long long*)(src+i+24)); 
                lb[4].i=*((long long*)(src+i+32)); 
                lb[5].i=*((long long*)(src+i+40)); 
                lb[6].i=*((long long*)(src+i+48)); 
                lb[7].i=*((long long*)(src+i+56)); 
                for(j=0;j<64 && i+j<rc;j++) 
                {       /* we let the compiler optimize this loop...
                        branch prediction should kill most the overhead...  */
                        *(dst+ri)=lb[0].b[j];  /* write byte in dst */
                        ri+=rl; /* next position in dst */
                        if(ri>=rc)ri-=rc-1;  /* go around and add one. */
                }
        }
}

/* decode buffer and write to fpout */

void decode_and_write(char* ec_buf,int rs_bsize,int bep_size,int rs_dsize,  int *newzeropositions, long* sum_of_failures,
                             long* uncorrectable_failures,char* argv,int quiet,int verbose, FILE *fpout)
{
        register int i, j;
        int nr_erasures;
        int eras_pos[255];

        for(i=0;i<bep_size;i++)
        {
                /* get the erasure positions and their number */
                nr_erasures =0;
                for (j=0; j < rs_bsize ; j++)
                {
                    if (newzeropositions[i*rs_bsize + j] != -1)
                    {
                         eras_pos[nr_erasures++]= newzeropositions[i*rs_bsize+j] ;
                    }
                }
                printf("ronde %d  nr_erasures %d \n", i, nr_erasures);

       
                int failure=rsd32((ec_buf+i*rs_bsize),eras_pos, nr_erasures);
                if (failure>0)
                {
                     printf("ronde %d failure corrected %d \n", i, failure);
                        *sum_of_failures+=failure;
                  if(!quiet && verbose) 
                  fprintf(stderr,"ronde %d %s: errors corrected: %ld uncorrectable blocks: %ld\r",
                                                      i, argv,*sum_of_failures,*uncorrectable_failures);
                }
                else if (failure==-1)
                {       
                       printf("uncorrectable failure \n");
                        *uncorrectable_failures++;
                        if(!quiet && verbose) 
                    fprintf(stderr,"ronde %d %s: errors corrected: %ld uncorrectable blocks: %ld\r",
                                                     i, argv,*sum_of_failures,*uncorrectable_failures);
                }
                if (rs_dsize!=fwrite((ec_buf+i*rs_bsize), sizeof(dtype),rs_dsize, fpout))
                {       /* write decoded data */
                        perror("serious problem: ");
                        exit(-1);
                }
        }
}                       


/*  main-program: reads commandline args and does anything else... */

int main(int argc,char *argv[])
{
  dtype *ec_buf = NULL;  /* pointer to encoding/decoding buffer */
  dtype *tr_buf = NULL;  /* pointer to transmit-buffer (fread/fwrite) */
  dtype *tr1_buf = NULL;  /* pointer to transmit-buffer (fread/fwrite) */
  dtype *bk_buf = NULL; /* pointer to backup-buffer for resync */
  long verbose=0,decode=0,quiet=0;
  int rs_bsize=0,rs_dsize=0,bep_size=0;
  long sum_of_failures=0;
  long uncorrectable_failures=0;
  int i,j, k;
  FILE *fpin, *fpout, *fperasures; 
  long got;
  int coded_file_size ;
  char coded_file_ext[4] ;
  char orgfilename[100];
  char *p ;
  int totsegs, seglength, startofseg, row;
  int *zeropositions, *newzeropositions;
  int *pointzero ;
  int nr_missing;
  int missingsegs[1000] ;
  int mis_seg ;

  if (argc != 2)
  {
	  printf("syntax is %s <filename.rs2>  \n", argv[0]) ;
	  exit(-1) ;
  }


  assert(sizeof(dtype)==1); /* anything works only with bytes */
  
  if ((fpin=fopen(argv[1], "rb"))==NULL)
  {
	  printf("Cannot open %s for input \n", argv[1]);
	  exit(1);
  }

  if ((fpout=fopen("tmpout.dat", "wb"))==NULL)
  {
	  printf("Cannot open tempout.dat for output \n");
	  exit(1);
  }
   
   tr1_buf=malloc(RS_BSIZE*MAX_EXPECTED_CHUNKS*sizeof(dtype)); 
   if (tr1_buf == NULL)
   {
	  fprintf(stderr, "cannot malloc tr_buf \n");
	 exit(-1);
   } 
   got =fread(tr1_buf,1, RS_BSIZE*MAX_EXPECTED_CHUNKS , fpin);

  bep_size = got/255 ;
  fseek(fpin, 0L, SEEK_SET); 

  if (got % 255)
	  bep_size++ ;
  decode = 1 ;  

        /* get the buffers we need (here we know how large): */
	  
	rs_bsize=RS_BSIZE;rs_dsize=RS_DSIZE;
        ec_buf=malloc(rs_bsize*bep_size*sizeof(dtype));  
        tr_buf=malloc(rs_bsize*bep_size*sizeof(dtype));  
        bk_buf=malloc(rs_bsize*bep_size*sizeof(dtype));  
        if(ec_buf==NULL || tr_buf==NULL || bk_buf==NULL)
        {
              fprintf(stderr,"%s: out of memory, bailing.\n",argv[0]);
              exit(-1);
        }

       /* read data from file erasures.dat */
       if ((fperasures = fopen("erasures.dat", "r"))==NULL)
       {
         printf("Cannot find erasures.dat \n");
         exit(1);
       }

       fscanf(fperasures, "%d", &totsegs);
       fscanf(fperasures, "%d", &seglength);
       nr_missing=0;
       while ( fscanf(fperasures, "%d", &mis_seg) != EOF)
       {
           missingsegs[nr_missing++] = mis_seg;
       }
       printf("In rs4decoderas nr_missing is %d \n", nr_missing);
       if ( nr_missing*seglength > (rs_bsize-rs_dsize)*bep_size )
       {
           printf("too many missing segments\n");
           return 1;
       }
       zeropositions = malloc(seglength*(totsegs+1)*sizeof(int));
       if ( zeropositions == NULL )
       {
          printf("cannot malloc zeropositions \n");
          exit(1);
       }
       for (i=0; i < seglength*(totsegs+1) ; i++)
                 zeropositions[i] = -1;      
        newzeropositions = malloc(seglength*(totsegs+1)*sizeof(int));
       if ( newzeropositions == NULL)
       {
          printf("cannot malloc newzeropositions \n");
          exit(1);
       }
       /* now label the erasures positions */
       for (i= 0 ; i < nr_missing; i++)
       {
           startofseg = missingsegs[i]*seglength ;
           for (j=0; j < seglength ; j++)
           {
              row= (startofseg +j ) / bep_size;
              /* if ( row < rs_dsize)  */
              zeropositions[startofseg+j] = row; 
           }
        }

        /*  distribute version pa0mbo for the indexes */
       pointzero=newzeropositions;
       for (i=0; i < bep_size ; i++)
       {
          for (j=0; j < 255 ; j++)
          {
            *(pointzero++) = *(zeropositions + j*bep_size + i);
          }
       }


        /* debugging 

        for (i=0; i < bep_size; i++)
        {
             for (j=0; j < 255; j++)
             {
                if (newzeropositions[i*255+j] == -1 )
                {
                   printf("- ");
                }
                else 
                   printf("%d ", newzeropositions[i*255 +j] );
             }
             printf("\n");
        } end debugging */
        
       /* go on: */
        while (!feof(fpin))
        {
                got=fread(tr_buf,1,bep_size*rs_bsize,fpin); /* read the encoded data */
				if (!got) break;
                if (got < (bep_size*rs_bsize))
                {
					for (k=got; k < bep_size*rs_bsize; k++)
						  tr_buf[k]=0;
					printf("added %ld zeros", bep_size*rs_bsize - got);
                }

		distribute(tr_buf,ec_buf,bep_size,rs_bsize,1);
                /* debugging kijk wat er in ec_buf staat 
                for (i=0; i < bep_size; i++)
                {
                    for (j=0; j < 255; j++)
                    {
                       printf("%x ", ec_buf[i*255 + j]);
                    }
                    printf("\n");
                }  */
		decode_and_write(ec_buf,rs_bsize,bep_size,rs_dsize, newzeropositions, &sum_of_failures,&uncorrectable_failures,argv[0],
				 quiet,verbose, fpout);
        }
        if((sum_of_failures>0 || uncorrectable_failures>0 || verbose) && !quiet)
        {
                fprintf(stderr,"\n%s: number of corrected failures   : %ld\n",argv[0],sum_of_failures);
                fprintf(stderr,"%s: number of uncorrectable blocks : %ld\n",argv[0],uncorrectable_failures);
        }
  /* clean up */
  fclose(fpin);
  fclose(fpout);

  if(ec_buf) free(ec_buf);
  if(tr_buf) free(tr_buf);
  if(bk_buf) free(bk_buf);
  printf("Na de-allocatie van buffers \n");

  /* now remove header and trailer from tmpout.dat 

   tr_buf=malloc(RS_BSIZE*MAX_EXPECTED_CHUNKS*sizeof(dtype)); 
   if (tr_buf == NULL)
   {
	  fprintf(stderr, "cannot malloc tr_buf \n");
	 exit(-1);
   } 
   printf("Na nieuwe allocatie van tr_buf \n");
   exit(0);  */
  if ((fpin=fopen("tmpout.dat", "rb"))==NULL)
  {
	  printf("Cannot open tempout.dat for input \n");
	  exit(1);
  }
  /* now read the tmpout data again */

  got =fread(tr1_buf,1, RS_BSIZE*MAX_EXPECTED_CHUNKS , fpin);
  
  if (bep_size != ((int) tr1_buf[1] + ((int) tr1_buf[2])*256 ))
  {
	  fprintf(stderr, "problems with bep_size coded in file \n");
	  printf("bep_size %d coded size %d \n", bep_size, (int) tr_buf[1] + ((int) tr_buf[2])*256) ;
		  exit(-1);
  }
  coded_file_size = bep_size*RS_DSIZE - (int)tr1_buf[0];
  strcpy(coded_file_ext, tr1_buf+3);
  strcpy(orgfilename, argv[1]);
  p = strtok(orgfilename, ".");
  strcat(orgfilename, ".");
  strcat(orgfilename, coded_file_ext);

  printf("newfilename: %s \n", orgfilename);

  if ((fpout=fopen(orgfilename, "wb"))==NULL)
  {
	  printf("Cannot open tempout.dat for output \n");
	  exit(1);
  }
  fwrite(tr1_buf+7,1, coded_file_size, fpout);
  fclose(fpout);

  /* now save erasures.dat under new name 
  strcpy(coded_file_ext, "era");
  strcpy(orgfilename, argv[1]);
  p = strtok(orgfilename, ".");
  strcat(orgfilename, ".");
  strcat(orgfilename, coded_file_ext);
  if ((fpout=fopen(orgfilename, "w"))==NULL)
  {
       printf("Cannot open %s for output\n", orgfilename);
       exit(1);
  }
  fprintf(fpout, "%d\n", totsegs);
  fprintf(fpout, " %d\n", seglength);
  for (i=0; i < nr_missing ; i++)
  {
     fprintf(fpout, " %d\n", missingsegs[i]);
  }
  fclose(fpout);  
  */

  return 0;
}

